"""Metadata pertaining to the xxkcd Python package"""

__author__ = 'Mital Ashok'
__credits__ = ['Mital Ashok']
__license__ = 'MIT'
__version__ = '1.4.0'
__maintainer__ = 'Mital Ashok'
__author_email__ = __email__ = 'mital.vaja@googlemail.com'
__status__ = 'Production'
__url__ = 'https://github.com/MitalAshok/xxkcd'
__description__ = 'An (unofficial) Python wrapper around xkcd APIs'
